
/* ******************************************************************************* */
/*                                                                                 */
/*  Practica de Intelligencia Artificial II                                       */
/*  Implementacio de l'algorisme de classificacio de dades multidimensionals:      */
/*  SOM (Self Oriented Map)                                                        */   
/*  Classe LecturaFitxer: conte la funcionalitat de lectura de patrons de fitxer   */
/*  Autors: Francisco Jose Aguilar Celdran i Sergio Blanco Cuaresma                */
/*                                                                                 */
/* ******************************************************************************* */


import java.util.*;
import java.io.*;

public class LecturaFitxer
{

   private float rangMenor[];
   private float rangMajor[];
   private Collection classes;
   private boolean classificat = false;

   public Collection llegirPatronsFitxer(FileReader arxiu, int tamPatrons, int colsClasses) throws ExcepcioFitxer
   {
      try
      {
         Collection patrons = new Vector();
         classes = new Vector();
         BufferedReader entrada = new BufferedReader(arxiu);
         String linia = entrada.readLine();
         int numPatro = 0;
         while (linia != null) {
            StringTokenizer valor = new StringTokenizer(linia);
            Float patro[] = new Float[tamPatrons];
            // Obtencio dels 'tamPatrons' de cada patro
            int comptador = 0;
            while (valor.hasMoreTokens() && comptador < tamPatrons) {
               String lectura = (String)valor.nextToken(); //System.out.println(lectura);
               patro[comptador] = new Float(Float.parseFloat(lectura));
               comptador++;
            }
            // Obtencio de la classe de cada patro, les darreres 'colsClasses' columnes
            comptador = 0;
            boolean trobada = false;
            while (valor.hasMoreTokens() && comptador < colsClasses && !trobada) {
               String lectura = (String)valor.nextToken(); //System.out.println(lectura);
               if ((Float.parseFloat(lectura)) == 1.0) { 
                  classes.add(new Float(comptador + 1)); //System.out.println("Classe trobada" + String.valueOf(comptador + 1));
                  trobada = true;
               }
               comptador++;
            }
            patrons.add(patro);
            linia = entrada.readLine();
            numPatro++;
         }
         arxiu.close();
         classificat = true;
         return patrons;
      }
      catch(Exception ex)
      {
         throw new ExcepcioFitxer("Error en la lectura del fitxer de dades: " + ex.getMessage());
      }
   }

   private void obtenirRangsDades(Collection rangs, int tamPatrons)
   {
      rangMenor = new float[tamPatrons];
      rangMajor = new float[tamPatrons];
      int comptador = 0;
      Iterator it = rangs.iterator();
      while (it.hasNext()) {
         Hashtable ht = (Hashtable)it.next();
         Integer menor = (Integer)ht.get("menor");
         rangMenor[comptador] = menor.intValue();
         Integer major = (Integer)ht.get("major");
         rangMajor[comptador] = major.intValue();
      }
   }
   
   public Collection llegirPatronsFitxer(FileReader arxiu, int tamPatrons, Collection rangs) throws ExcepcioFitxer
   {
      try
      {
         obtenirRangsDades(rangs,tamPatrons);
         Collection patrons = new Vector();
         BufferedReader entrada = new BufferedReader(arxiu);
         String linia = entrada.readLine();
   
         while (linia != null) {
            StringTokenizer valor = new StringTokenizer(linia);
            Float patro[] = new Float[tamPatrons];
            int comptador = 0;
            while (valor.hasMoreTokens() && comptador < tamPatrons) {
               String lectura = (String)valor.nextToken();
               float dada = Float.parseFloat(lectura);
               dada -= rangMenor[comptador];
               float rang = rangMajor[comptador] - rangMenor[comptador];
               dada /= rang;
               patro[comptador] = new Float(dada);
               comptador++;
            }
            patrons.add(patro);
            linia = entrada.readLine();
         }
         arxiu.close();
         classificat = false;
         return patrons;
      }
      catch(Exception ex)
      {
         throw new ExcepcioFitxer("Error en la lectura del fitxer de dades");
      }
   }
   
   public Collection obtenirClasses()
   {
      if (classificat) {
         return classes;
      }
      else {
         return null;
      }
   }

}

