
/* ******************************************************************************* */
/*                                                                                 */
/*  Practica de Intelligencia Artificial II                                       */
/*  Implementacio de l'algorisme de classificacio de dades multidimensionals:      */
/*  SOM (Self Oriented Map)                                                        */   
/*  Classe SOM: conte la logica de l'aplicacio                                     */
/*  Autors: Francisco Jose Aguilar Celdran i Sergio Blanco Cuaresma                */
/*                                                                                 */
/* ******************************************************************************* */


import java.util.*;
import java.io.*;


public class Som
{
   private float[][][] malla;
   private float[][] patrons;
   private int numPatrons;
   private int dimX;
   private int dimY;
   private int tamPatro;
   private int numEpoques;
   private int radiVeinatge;
   private int millorX;
   private int millorY;
   private float taxaAprenentatge;
   private float sigma;
   private int colsClasses;
   private int multiplicadorEpoques;
   private boolean dadesClassificades;
   private Collection rangsPatrons;
   private Collection classes;   
   private Collection dadesFitxer;
   private boolean lecturaFitxer;
   private float[][] mallaClassificada;
   private Collection[][] patronsClassificats;
   
   
   // Funcio constructora que inicialitza els parametres del programa 
   
   public Som(int dimX, int dimY, int tam, int numEpoques, int classes, 
               Collection rangs, int multiplicador, float taxaAprenentatge, 
               float sigma, int radi)
   {
      this.dimX = dimX;
      this.dimY = dimY;
      tamPatro = tam;
      this.numEpoques = numEpoques;
      radiVeinatge = radi;
      colsClasses = classes;
      rangsPatrons = rangs;
      multiplicadorEpoques = multiplicador;
      if (colsClasses == 0) {
         dadesClassificades = false;
      }
      else {
         dadesClassificades = true;
      }
      this.sigma = sigma;
      this.taxaAprenentatge = taxaAprenentatge;
      lecturaFitxer = false;
   }

   // Lectura i emmagatzemament dels patrons del fitxer especificat
   
   public void obtenirPatrons(File fitxerDades) throws ExcepcioFitxer
   {
      try
      {
         LecturaFitxer lectura = new LecturaFitxer();
         FileReader entrada = new FileReader(fitxerDades);
         if (dadesClassificades) {
            dadesFitxer = lectura.llegirPatronsFitxer(entrada,tamPatro,colsClasses);
            classes = lectura.obtenirClasses();             
         }
         else {
            dadesFitxer = lectura.llegirPatronsFitxer(entrada,tamPatro,rangsPatrons);
         }
         Iterator it = dadesFitxer.iterator();
         numPatrons = dadesFitxer.size();
         patrons = new float[numPatrons][tamPatro];
         int comptador = 0;
         while(it.hasNext()) {
            Float patro[] = (Float [])it.next();
            for (int i = 0; i < patro.length; i++) {
               patrons[comptador][i] = patro[i].floatValue();
            }
            comptador++;
         }
         lecturaFitxer = true;
      }
      catch(FileNotFoundException ex)
      {
         throw new ExcepcioFitxer("No s'ha pogut localitzar el fitxer de patrons");
      }     
   }

   // Inicialitzacio dels valors dels pesos de la malla, aleatoriament 
      
   private void inicialitzacioAleatoriaPesos()
   {
      float aleatori;
      Date data = new Date();
      Random generador = new Random();
      generador.setSeed(data.getTime());
      for (int i = 0; i < dimX; i++) {
         for (int j = 0; j < dimY; j++) {
            for (int k = 0; k < tamPatro; k++) {
               aleatori = generador.nextFloat();
               malla[i][j][k] = aleatori;
            }
         }
      }
   }

   // Algorisme de desordenacio de la taula de patrons, retornant-la desordenada

   private float[][] desordenarTaulaPatrons(float ordenats[][])
   {
      boolean fets[];
      float desordenats[][];
      int comptador = 0;
      
      fets = new boolean[numPatrons];
      for (int i = 0; i < numPatrons; i++) {
         fets[i] = false;
      }
      desordenats = new float[numPatrons][tamPatro];
      Date data = new Date();
      Random generador = new Random();
      generador.setSeed(data.getTime());
      while (comptador < numPatrons) {
         int aleatori = generador.nextInt(numPatrons);
         if (!fets[aleatori]) {
            desordenats[comptador] = ordenats[aleatori];
            fets[aleatori] = true;
            comptador++;
         }
      }
      return desordenats;
   }

   
   private float[][] desordenarTaulaPatrons2(float ordenats[][])
   {
      boolean fets[];
      float desordenats[][];
      int comptador = 0;
      
      fets = new boolean[numPatrons];
      for (int i = 0; i < numPatrons; i++) {
         fets[i] = false;
      }
      desordenats = new float[numPatrons][tamPatro];
      Date data = new Date();
      Random generador = new Random();
      generador.setSeed(data.getTime());
      while (comptador < numPatrons) {
         int aleatori = generador.nextInt(numPatrons);
         if (!fets[aleatori]) {
            desordenats[comptador] = ordenats[aleatori];
            fets[aleatori] = true;
            comptador++;
         }
      }
      return desordenats;
   }
   
   // Calcul de la distancia entre el vector de pesos de la neurona actual
   // i el vector de valors del patro actual

   private float calcularDistancia(float neurona[], float patro[])
   {
      float distancia = 0;
      for (int i = 0; i < tamPatro; i++)  {
         distancia += (float)Math.pow((patro[i] - neurona[i]),2);
      }
      return distancia;
   }

   // Calcul de la neurona amb menor distancia respecte al patro actual

   private void calcularNeuronaGuanyadora(float patro[])
   {
      float distanciaMenor = Float.MAX_VALUE;
      
      for(int i = 0; i < dimX; i++) {
         for(int j = 0; j < dimY; j++) {
            float distanciaActual = calcularDistancia(malla[i][j],patro);
            if (distanciaActual < distanciaMenor) {
               distanciaMenor = distanciaActual;
               millorX = i;
               millorY = j;
            }
         }
      }
   }

   // Mesura de la distancia del vector normalitzat, aplicat a la diferencia de: 
   // la neurona de la malla amb coordenades posX, posY i la neurona guanyadora 
   // amb coordenades millorX, millorY   || r(i) - r(c) ||
   
   private float distanciaMalla(int posX, int posY)
   {
      float distancia = 0;
      float vector[] = new float[tamPatro];
      for(int i = 0; i < tamPatro; i++) {
         vector[i] = malla[posX][posY][i] - malla[millorX][millorY][i];
         distancia += (float)Math.pow(vector[i],2);
      }
      distancia = (float)Math.sqrt(distancia);
      return distancia;
   }
   
   // Obtencio del valor resultant d'aplicar la funcio de veinatge a la neurona actual
   
   private float funcioVeinatge(int posX, int posY, float patro[])
   {
      float resultat = 0;
      float numerador = distanciaMalla(posX,posY);
      numerador = (float)Math.pow(numerador,2);
      float denominador = 2 * (float)Math.pow(sigma,2);
      float exponent = - (float)numerador / denominador; 
      float veinatge = (float)Math.pow(Math.E,exponent);
      resultat = taxaAprenentatge * veinatge;
      return resultat;
   }

   // Modificacio del pes d'una neurona de la malla dins el radi de veinatge
   // de la neurona guanyadora, segons els passos de la formula de l'algorisme SOM

   private void modificarPesNeurona(int x, int y, int radi, float patro[])
   {
      float pes[] = new float[tamPatro];
      
      for(int i = 0; i < tamPatro; i++) {
         pes[i] = malla[x][y][i];
         float diferencia = patro[i] - malla[x][y][i];
         pes[i] = pes[i] + (funcioVeinatge(x,y,patro) * diferencia);
         malla[x][y][i] = pes[i];
      }
   }

   // Funcio que calcula A modul N

   private int modul(int a, int n) 
   {
      int mod;
      mod = a - (a/n) * n;
      if (mod < 0) {
         mod = n + mod;
      } 
      return mod;
   }

   // Calcul de la distancia entre dues posicions de la malla tenint en compte
   // condicions de contorn periodiques
   
   public int diferenciaPosicions(int actual, int millor, int mod)
   {
      int menor;
      int distEsq;
      int distDrt;
      
      if (millor > actual) {
         distEsq = millor - actual;
         distDrt = modul((actual - millor),mod);
      }
      else {
         distEsq = actual - millor;
         distDrt = modul((millor - actual),mod);
      }
      menor = Math.min(distEsq,distDrt);     
      return menor;
   }

   // Modificacio dels pesos de la neurona guanyadora i de les neurones
   // del radi indicat per parametre

   private void modificarPesos(int radi, float patro[])
   {
      int distX = modul((millorX - radi),dimX);
      int distY = modul((millorY - radi),dimY);
      int comptadorX = (2 * radi) + 1;
            
      while (comptadorX >= 0) {
         distY = modul((millorY - radi),dimY);
         int comptadorY = (2 * radi) + 1;
         while (comptadorY >= 0) {
            int distanciaX = diferenciaPosicions(distX,millorX,dimX);
            int distanciaY = diferenciaPosicions(distY,millorY,dimY);
            int total = distanciaX + distanciaY;
            if (total <= radi) { 
               modificarPesNeurona(distX,distY,radi,patro);
            }
               distY = modul((distY + 1),dimY);
            comptadorY--;
         }
            distX = modul((distX + 1),dimX);
         comptadorX--;
      }
   }

   // Calcul de la distancia mitjana entre la neurona guanyadora i tots els patrons

   private float distanciaMitjaGuanyadores()
   {
      float distMitja = 0;
      for (int i = 0; i < numPatrons; i++) {
         calcularNeuronaGuanyadora(patrons[i]);
         distMitja += calcularDistancia(malla[millorX][millorY],patrons[i]);
      }
      distMitja = (float)distMitja / numPatrons;
      return distMitja;    
   }

   // Programa principal de l'Algorisme Self Organizing Map (SOM)

   public boolean execucioAlgorismeSOM()
   {
      boolean primerCop = true;
      
      if (!lecturaFitxer) {
         return false;
      }

      malla = new float[dimX][dimY][tamPatro];
         
         
      // Generacio de pesos aleatoris inicials per la malla
      inicialitzacioAleatoriaPesos();
      
      float distanciaActual = 0;
      float distanciaAnterior = 0;
      float error = 0;
      float errorActual = 0;
      float errorAnterior = 0;
      int comptadorError = 0;
      int epoques = 0;
      int nombreDecrements = 0;
      
      // Execucio de l'algorisme SOM 'numEpoques' vegades,
      // reordenant en cada epoca l'ordre de la taula de patrons     
      while (epoques < numEpoques) {
         error = 0;
         for(int patroActual = 0; patroActual < numPatrons; patroActual++) {
            calcularNeuronaGuanyadora(patrons[patroActual]);
            error += calcularDistancia(malla[millorX][millorY],patrons[patroActual]);
            modificarPesos(radiVeinatge,patrons[patroActual]);
         }
         error = (float)Math.round( ((float)error / numPatrons) * 1000) /1000;
         
         switch (epoques) {
            case 0:
               errorActual = error;
               comptadorError++;
               break;
            default:
               comptadorError++;
               if ((comptadorError == 10) && (nombreDecrements <= 38)) {
                  comptadorError = 0;
                  errorAnterior = errorActual;
                  errorActual = error;
                  // Si estabilitza l'error
                  if ((errorActual >= errorAnterior)) {
                     nombreDecrements++;
                     if (nombreDecrements <= 38) {
                        taxaAprenentatge = (float)taxaAprenentatge / (float)1.2;
                        sigma = (float)sigma / (float)1.2;
                     }
                  }
                  
               }
               break; 
         }
         patrons = desordenarTaulaPatrons(patrons);
         epoques++;
         
         // Es multiplica per 'multiplicadorEpoques' el nombre d'epoques
         // en arribar al nombre d'epoques de l'actual radi de veinatge.
         // Es disminueix el radi de veinatge, la taxa d'aprenentatge i la sigma
         if (epoques == numEpoques) {
            if (radiVeinatge > 1) {
               epoques = 0;
               numEpoques *= multiplicadorEpoques;
               radiVeinatge -= 1;
            }
         }
         
      }
      return true;
   }  
   
   // Classificacio de les dades una vegada entrenada la malla de neurones,
   // a partir de les neurones amb menor distancia respecte a cada patro. 
   // 'mallaClassificada' indica les posicions de la malla on s'han trobat
   // neurones guanyadores, especificant el nombre de neurones guanyadores
   // en aquesta posicio, o, en cas de tenir les classes previament definides,
   // a quina classe de patro correspon la neurona guanyadora d'aquella posicio.
   // 'patronsClassificats' indica a quin patro o conjunt de patrons corresponen
   // les neurones guanyadores de cada posicio de la malla
   
   public Hashtable classificarDades() 
   {
      Iterator it;
      
      mallaClassificada = new float[dimX][dimY];
      patronsClassificats = new Collection[dimX][dimY];
      for (int i = 0; i < dimX; i++) {
         for (int j = 0; j < dimX; j++) {
            mallaClassificada[i][j] = 0;
            patronsClassificats[i][j] = new Vector();
         }
      }
      float taulaClasses[] = new float[numPatrons];
      if (dadesClassificades) {
         it = classes.iterator();
         int comptador = 0;
         while (it.hasNext()) {
            taulaClasses[comptador] = ((Float)it.next()).floatValue();
            comptador++;
         }
      }
      it = dadesFitxer.iterator();
      int comptador = 0;
      while (it.hasNext()) {
         Float patro[] = (Float [])it.next();
         float patrof[] = new float[tamPatro];
         for (int i = 0; i < patro.length; i++) {
               patrof[i] = patro[i].floatValue();
         }
         calcularNeuronaGuanyadora(patrof);
         if (dadesClassificades) {
            mallaClassificada[millorX][millorY] = taulaClasses[comptador];
         }
         else {
            mallaClassificada[millorX][millorY]++;
         }
         patronsClassificats[millorX][millorY].add(new Integer(comptador));
         comptador ++;
      }
      Hashtable ht = new Hashtable();
      ht.put("malla",mallaClassificada);
      ht.put("patrons",patronsClassificats);
      return ht;  
   }
   
   // Emmagatzemament en fitxer dels pesos de la malla entrenada per a 
   // posterior aplicacio en classificacio de patrons
   
   public void emmagatzemarMalla(String nomFitxer)
   {
      try {
         File f = new File(nomFitxer);
         FileWriter fw = new FileWriter(f);
         BufferedWriter buffer = new BufferedWriter(fw);
         PrintWriter sortida = new PrintWriter(buffer);
         sortida.println("Dimensio_X_Malla: " + String.valueOf(dimX));
         sortida.println("Dimensio_Y_Malla: " + String.valueOf(dimY));
         sortida.println("Nombre_pesos_per_neurona: " + String.valueOf(tamPatro));
         for (int i = 0; i < dimX; i++) {
            for (int j = 0; j < dimX; j++) {
               for (int k = 0; k < tamPatro; k++) {
                  sortida.print(malla[i][j][k] + " ");
               }
               sortida.print(String.valueOf(mallaClassificada[i][j]));
               sortida.println();
            }
         }
         sortida.close();
      }
      catch (Exception ex) {}
   }
   
   // Extraccio del valor numeric de la segona 'paraula' de la linia rebuda
   
   private int extreureTokenLinia(String linia)
   {
      StringTokenizer separador = new StringTokenizer(linia);
      separador.nextToken();
      int token = Integer.parseInt(separador.nextToken());
      return token;
   }
   
   // Classificacio dels patrons del 'fitxerPatrons' en funcio de la malla
   // previament entrenada i emmagatzemada en el 'fitxerMalla'
   
   public void aplicarMallaEntrenada(String fitxerMalla, String fitxerPatrons,
                                      String sortida)
   {
      try {
         File f = new File(fitxerPatrons);
         colsClasses = 0;
         obtenirPatrons(f);
         f = new File(fitxerMalla);
         FileReader entrada = new FileReader(f);
         BufferedReader buffer = new BufferedReader(entrada);
         dimX = extreureTokenLinia(buffer.readLine());
         dimY = extreureTokenLinia(buffer.readLine());
         tamPatro = extreureTokenLinia(buffer.readLine());
         float mallaClassificada[][] = new float[dimX][dimY];
         malla = new float[dimX][dimY][tamPatro];
         String linia = buffer.readLine();
         int comptadorX = 0;
         int comptadorY = 0;
         while (linia != null) {
            StringTokenizer valor = new StringTokenizer(linia);
            int numPes = 0;
            while (valor.hasMoreTokens() && numPes < tamPatro) {
               String lectura = (String)valor.nextToken();
               malla[comptadorX][comptadorY][numPes] = Float.parseFloat(lectura);
               numPes++;
            }
            String lectura = (String)valor.nextToken();
            mallaClassificada[comptadorX][comptadorY] = Float.parseFloat(lectura);
            if (comptadorY == (dimY - 1)) {
               comptadorX++;
               comptadorY = 0;
               if (comptadorX == dimX) {
                  break;
               }
            }
            else {
               comptadorY++;
            }
            linia = buffer.readLine();
         }
         buffer.close();
         File fitxerSortida = new File(sortida);
         FileWriter fw = new FileWriter(fitxerSortida);
         BufferedWriter bufferOut = new BufferedWriter(fw);
         PrintWriter fluxSortida = new PrintWriter(bufferOut);
         for (int i = 0; i < numPatrons; i++) {
            calcularNeuronaGuanyadora(patrons[i]);
            float classe = mallaClassificada[millorX][millorY];
            fluxSortida.print("Patro " + i + " (");
            for (int j = 0; j < tamPatro; j++) {
               fluxSortida.print(" " + patrons[i][j]);
            }
            fluxSortida.print(" ) --> Classe: ");
            fluxSortida.print(classe);
            fluxSortida.println();
         }
         fluxSortida.close();
         
      }
      catch(Exception ex) {}
   }
   
}

